Package: illuminaio
Version: 0.54.0
Title: Parsing Illumina Microarray Output Files
Imports: base64
Suggests: RUnit, BiocGenerics, IlluminaDataTestFiles (>= 1.0.2),
        BiocStyle
Description: Tools for parsing Illumina's microarray output files, including IDAT.	
Authors@R: c(person("Keith", "Baggerly", role = "aut"),
             person("Henrik", "Bengtsson", role = "aut"),
             person(c("Kasper", "Daniel"), "Hansen", role = c("aut", "cre"),
                    email = "kasperdanielhansen@gmail.com"),
             person("Matt", "Ritchie", role = "aut"),
             person(c("Mike", "L."), "Smith", role = "aut"),
             person("Tim", "Triche Jr.", role = "ctb"))
License: GPL-2
URL: https://github.com/HenrikBengtsson/illuminaio
BugReports: https://github.com/HenrikBengtsson/illuminaio/issues
biocViews: Infrastructure, DataImport, Microarray, ProprietaryPlatforms
git_url: https://git.bioconductor.org/packages/illuminaio
git_branch: RELEASE_3_23
git_last_commit: 86b198e
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 22:42:02 UTC; biocbuild
Author: Keith Baggerly [aut],
  Henrik Bengtsson [aut],
  Kasper Daniel Hansen [aut, cre],
  Matt Ritchie [aut],
  Mike L. Smith [aut],
  Tim Triche Jr. [ctb]
Maintainer: Kasper Daniel Hansen <kasperdanielhansen@gmail.com>

importFrom(utils, file_test)
importFrom(utils, packageDescription)
importFrom(utils, read.delim)
importFrom(utils, write.table)
importFrom(base64, decode)
useDynLib(illuminaio)
export(readBGX)
export(readBPM)
export(readIDAT)
useDynLib("illuminaio", .registration = TRUE, .fixes = "C_")

readIDAT <- function(file, what = c("all", "IlluminaID", "nSNPsRead")) {

    ## Wrapper function to determine IDAT format and call appropriate reading routine.
    ## Currently this just checks the magic "IDAT" string and then reads the version number
    ## The file name is then passed and the file opened again by the read function.
    
    stopifnot(is.character(file) || length(file) != 0)
    file <- path.expand(file)
    if(!file.exists(file)) {
        stop("Unable to find file ", file)
    }
    
    if(grepl("\\.gz", file))
        con <- gzfile(file, "rb")
    else
        con <- file(file, "rb")
    on.exit({
        if (!is.null(con)) close(con)
    })
    what <- match.arg(what)
    
    ## Assert file format
    magic <- readChar(con, nchars=4)
    if (magic != "IDAT") {
        stop("Cannot read IDAT file. File format error. Unknown magic: ", magic)
    }
    
    ## Read IDAT file format version
    version <- readBin(con, what="integer", size=4, n=1, signed = TRUE, endian = "little")

    close(con)
    con <- NULL

    if (version == 1) {
        if(what != "all")
            stop("This file is encrypted. For encrypted files we need `what` equal to `all`.")
        res <- readIDAT_enc(file)
    }
    else if (version == 3) {
        res <- readIDAT_nonenc(file, what = what)
    }
    else {
        stop("Cannot read IDAT file. Unsupported IDAT file format version: ", version)
    }
    
    res
}

library(illuminaio)
library(IlluminaDataTestFiles)

idatFile <- system.file("extdata", "idat", "4343238080_A_Grn.idat", package = "IlluminaDataTestFiles")
idatData <- readIDAT(idatFile)$Quants
gsFile <- system.file("extdata", "gs", "4343238080_A_ProbeSummary.txt.gz", package = "IlluminaDataTestFiles")
gStudio <- read.delim(gzfile(gsFile, open = "r"), sep = "\t", header = TRUE)

## not all probes are present in GenomeStudio output, so only select those that are
idatData <- idatData[which(idatData[,"CodesBinData"] %in% gStudio[,"ProbeID"]),]
## the orders are also different (numeric vs alphabetical)
gStudio <- gStudio[match(idatData[,"CodesBinData"], gStudio[,"ProbeID"]),]
# check each value in GenomeStudio output
## there are some rounding differences, so we allow slight differences

## summarised bead intensities
RUnit::checkEqualsNumeric(idatData[, "MeanBinData"], gStudio[, 3], tolerance = 10e-7)
## number of beads
RUnit::checkEquals(idatData[, "NumGoodBeadsBinData"], gStudio[, 5])
## standard errors
RUnit::checkEqualsNumeric(idatData[, "DevBinData"] / sqrt(idatData[, "NumGoodBeadsBinData"]), gStudio[, 4], tolerance = 10e-7)

library(illuminaio)
library(IlluminaDataTestFiles)

load(file.path(path.package("IlluminaDataTestFiles"), "extdata", "testData", "idat.ht12v4.rda"))
idatFile <- system.file("extdata", "idat", "6016741005_A_Grn.idat",
                        package = "IlluminaDataTestFiles")
idat <- illuminaio::readIDAT(idatFile)
RUnit::checkEquals(idat, idat.ht12v4)

library(illuminaio)
library(IlluminaDataTestFiles)

idatFile1 <- system.file("extdata", "idat", "4019585376_B_Red.idat",
                         package = "IlluminaDataTestFiles")
idat1 <- illuminaio::readIDAT(idatFile1)
idatFile2 <- system.file("extdata", "idat", "4019585376_B_Red.idat.gz",
                         package = "IlluminaDataTestFiles")
idat2 <- illuminaio::readIDAT(idatFile2)
idat2$fileSize <- idat1$fileSize <- NULL
RUnit::checkEquals(idat1, idat2)

library(illuminaio)
library(IlluminaDataTestFiles)

load(file.path(path.package("IlluminaDataTestFiles"), "extdata", "testData", "idat.wg6v2.rda"))
idatFile <- system.file("extdata", "idat", "4343238080_A_Grn.idat",
                        package = "IlluminaDataTestFiles")
idat <- illuminaio::readIDAT(idatFile)
RUnit::checkEquals(idat, idat.wg6v2)

library(IlluminaDataTestFiles)

load(file.path(path.package("IlluminaDataTestFiles"), "extdata", "testData", "idat.450k.rda"))
idatFile <- system.file("extdata", "idat", "5723646052_R02C02_Grn.idat",
                        package = "IlluminaDataTestFiles")
idat <- illuminaio::readIDAT(idatFile)

## WORKAROUND: Until 'IlluminaDataTestFiles' is updated, 
## don't test all of the "Unknowns" fields
if (  "Unknown.7" %in% names(idat[["Unknowns"]]) &&
    ! "Unknown.7" %in% names(idat.450k[["Unknowns"]])) {
  unknowns <- idat[["Unknowns"]]
  unknowns[["Unknown.6"]] <- NULL
  unknowns[["Unknown.7"]] <- NULL
  idat[["Unknowns"]] <- unknowns
}

RUnit::checkEquals(idat, idat.450k)

library(illuminaio)
library(IlluminaDataTestFiles)

load(file.path(path.package("IlluminaDataTestFiles"), "extdata", "testData", "idat.370k.rda"))
idatFile <- system.file("extdata", "idat", "4019585376_B_Red.idat",
                        package = "IlluminaDataTestFiles")
idat <- illuminaio::readIDAT(idatFile)
## FIXME: IlluminaDataTestFiles needs to be updated according
##        to 'RunInfo' bug fix in illuminaio v0.11.2.
if (packageVersion("illuminaio") > "0.11.1") idat$RunInfo[-(1:2),] <- NA
RUnit::checkEquals(idat, idat.370k)

readBPM <- function(file) {
    .Deprecated(msg="readBPM() is deprecated since illuminaio 0.1.4 (January 2016), because its parser is invalid and broken. (Issue #6)")

    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ## Local functions
    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    readByte <- function(con, n=1L, ...) {
      readBin(con, what="integer", n=n, size=1L, endian="little", signed=FALSE)
    }

    readShort <- function(con, n=1L, ...) {
      readBin(con, what="integer", n=n, size=2L, endian="little", signed=FALSE)
    }

    readInt <- function(con, n=1L, ...) {
      readBin(con, what="integer", n=n, size=4L, endian="little", signed=TRUE)
    }

    readLong <- function(con, n=1L, ...) {
      readBin(con, what="integer", n=n, size=8L, endian="little", signed=TRUE)
    }

    readString <- function(con, ...) {
      n <- readByte(con, n=1L)
      readChar(con, nchars=n)
    }

    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ## Validate arguments
    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ## Argument 'file':
    file <- path.expand(file)
    if (!file_test("-f", file)) {
        stop("File not found: ", file)
    }


    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ## Open file (and make sure it's closed at the end)
    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    con <- file(file, open="rb")
    on.exit({
        close(con)
    })


    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ## Reads and parses Illumina BPM files
    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    ## The first few bytes of the egtFile are some type of
    ## header, but there's no related byte offset information.

    ## Assert file format
    prefixCheck <- readChar(con, nchars=3L) ## should be "BPM"
    if (prefixCheck != "BPM") {
       stop("Cannot read BPM file. File format error. Unknown magic: ", prefixCheck)
    }

    null.1 <- readByte(con)
    ## should be 1

    versionNumber <- readInt(con, n=1L)
    ## should be 4
    if (versionNumber != 4L) {
      stop("Cannot read BPM file. Unsupported BPM file format version: ", versionNumber)
    }

    chipType <- readString(con)

    null.2 <- readByte(con, n=2L)

    csvLines <- readLines(con, n=22L)

    entriesByteOffset <- seek(con)
    nEntries <- readInt(con, n=1L)
    if (!is.finite(nEntries) || nEntries < 0) {
       stop("Cannot read BPM file. File format error. Invalid number of entries to read: ", nEntries)
    }

    if (FALSE) {
        snpIndexByteOffset <- seek(con)
        snpIndex <- readInt(con, n=nEntries)
        ## for the 1M array, these are simply in order from 1 to 1072820.

        snpNamesByteOffset <- seek(con)
        snpNames <- rep("A", times=nEntries)
        for(i1 in 1:nEntries){
            snpNames[i1] <- readString(con)
        }
    }

    seek(con, where=15278138L)

    normIDByteOffset <- seek(con)
    normID <- readByte(con, n=nEntries) + 1L

    newBlockByteOffset <- seek(con)
    newBlock <- readByte(con, n=10000L)


    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ## Return results
    ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    byteOffsets <- list(entriesByteOffset=entriesByteOffset,
                        ##snpIndexByteOffset=snpIndexByteOffset,
                        ##snpNamesByteOffset=snpNamesByteOffset,
                        normIDByteOffset=normIDByteOffset,
                        newBlockByteOffset=newBlockByteOffset)

    list(prefixCheck=prefixCheck,
         null.1=null.1,
         versionNumber=versionNumber,
         chipType=chipType,
         null.2=null.2,
         csvLines=csvLines,
         nEntries=nEntries,
         ##snpIndex=snpIndex,
         ##snpNames=snpNames,
         normID=normID,
         newBlock=newBlock,
         byteOffsets=byteOffsets)
} # readBPM()

readIDAT_enc <- function(file) {
    if(!file.exists(file)) {
        stop("Unable to find file ", file)
    }

    tempFile <- tempfile()
    
    #if(verbose) 
    #    message("Decrypting to XML")

    out <- .C("decrypt", as.character(file), as.character(tempFile), PACKAGE = "illuminaio")
    
    #if(verbose)
    #    message("Reading XML")
    
    r <- readLines(tempFile, warn = FALSE)
    file.remove(tempFile)   

    what = c(rep("numeric", 6), rep("integer", 4))
    data <- list()
    
    tf <- tempfile(c("", ""))

    ## if we've got the XML format with lots of short lines we're
    ## going to find the lines that indicate the start of a new entry
    ## and then build each block of data by concatanating the lines inbetween 
    if(length(r) > 100) {

        ## find the lines that contain the names for each block of data
        idx <- grep(" __", r)
        idx <- c(idx, grep(">", r)[2])

        for(i in 1:(length(idx)-1)) {
                     
            tmp <- strsplit(r[idx[i]], " __")[[1]][2]

            name <- strsplit(tmp, "=\\\"")[[1]][1]

            firstLine <- strsplit(tmp, "=\\\"")[[1]][2]
            middleLines <- paste(r[ (idx[i]+1):(idx[i+1]-1) ], collapse = "")
            finalLine <- strsplit(r[idx[i+1]], "\ __")[[1]][1]           
            b64String <- paste(firstLine, middleLines, finalLine, sep = "")
            
            ## remove quotation mark (and for last entry the closing angle bracket)
            if(i < length(idx)-1)
                b64String <- substr(b64String, 1, nchar(b64String) - 1)
            else
                b64String <- substr(b64String, 1, nchar(b64String) - 2)

            ## write base64 to file, then decode, then write binary
            write.table(file = tf[1], x = b64String, row.names = FALSE, col.names = FALSE, quote = FALSE)
            decode(input = tf[1], output = tf[2])

            ## estimated number of bead-types (plus a small fudge factor to be sure)
            nDataPoints <- (3 * nchar(b64String) / 16) + 100
            data[[name]] <- readBin(tf[2], what = what[i], size = 4, n = as.integer(nDataPoints))           
        }
    } else {
        ## if we've got one long line for all the data chunks then we split it
        ## based on the position of the "__" indicating the data chunk names
        ## then process each of those in turn
       
        dataChunks <- strsplit(r[2], " __")[[1]]
        for(i in 2:length(dataChunks)) {
            tmp <- strsplit(dataChunks[[i]], "\\\"")[[1]]
            name <- substr(tmp[1], 1, nchar(tmp[1]) - 1)
            b64String <- tmp[2]

            write.table(file = tf[1], x = b64String, row.names = FALSE, col.names = FALSE, quote = FALSE)
            base64::decode(input = tf[1], output = tf[2])

            nDataPoints <- (3 * nchar(b64String) / 16) + 100
            data[[name]] <- readBin(tf[2], what = what[i-1], size = 4, n = as.integer(nDataPoints))
        }
    }
    file.remove(tf[1], tf[2])
    
    ## I have found one instance where the entries in data[[]] are not all the same length.  
    ## This hack removes them, although this is probably not optimal
    dataLength <- unlist(lapply(data, length))
    commonLength <- as.integer(names(sort(table(dataLength), decreasing=TRUE)))[1]
    extra <- NULL
    if(any(dataLength != commonLength)) {
        extra <- list()
        for(i in which(dataLength != commonLength)) {
            extra[[ names(data)[i] ]] <- data[[i]]
            data[[i]] <- NULL
      }
    }
    

    ## grab some of the other information stored in the XML file
    runInfo <- extractRunInfo(r)
    chipInfo <- extractChipInfo(r)
    
    res <- list(
             Barcode=chipInfo$BarCode,
             Section=chipInfo$SectionLabel,
             ChipType=chipInfo$SentrixFormat,
             Quants=as.data.frame(data, stringsAsFactors=FALSE),
             RunInfo=runInfo
           )
    if(!is.null(extra)) 
        res[["Extra"]] <- extra
    
    res
    
}

extractRunInfo <- function(lines) {
    
    idx <- cbind(grep("<ProcessEntry>", lines), grep("</ProcessEntry>", lines))
    ## VeraCode data doesn't contain this run information, so we just return nothing
    if(!nrow(idx)) {
      res <- NULL
    } else {  
      fields <- c("Name", "SoftwareApp", "Version", "Date", "Parameters")
      res <- matrix(NA_character_, ncol = length(fields), nrow = nrow(idx), dimnames = list(rep("", nrow(idx)), fields))
      
      for(i in 1:nrow(idx)) {
          entry <- lines[idx[i,1]:idx[i,2]]
  
          for(f in fields) {
              line <- entry[grep(paste("<", f, ">", sep = ""), entry)]
              ## extract the string between tags
              start <- gregexpr(">", line)[[1]][1] + 1
              end <- gregexpr("<", line)[[1]][2] - 1
              res[i, f] <- substring(line, start, end)
          }
      }
    }
    
    res
}
        
extractChipInfo <- function(lines) {
    fields <- c("BarCode", "SentrixFormat", "SectionLabel")
    res <- list()
    
    for(f in fields) {
        line <- lines[grep(paste("<", f, ">", sep = ""), lines)]
        ## the field may not exist
        if(!length(line)) {
          start <- end <- NA
        } else {
          ## extract the string between tags
          start <- gregexpr(">", line)[[1]][1] + 1
          end <- gregexpr("<", line)[[1]][2] - 1
        }
        ## with VeraCode data these can be empty tags, so we'll return NULL in those cases
        if(!is.na(end)) {
          res[[ f ]] <- substring(line, start, end)
        } else {
          res[[ f ]] <- NULL
        }
    }
    
    res
}
    

readBGX <- function(file) {
 
    ## read the [Heading] section of the file (normally 7 lines, but we'll be a little cautious)
    tmp <- read.delim(gzfile(file), sep = "\t", header = FALSE, nrows = 10, stringsAsFactors = FALSE)

    ## get the number of probes and controls, and find the line the probes start on
    nProbes <- as.integer(tmp[grep("Number of Probes", tmp[,1]), 2])
    nControls <- as.integer(tmp[grep("Number of Controls", tmp[,1]), 2])
    probesHead <- grep("\\[Probes\\]", tmp[,1])

    ## read probes and then controls
    con <- gzfile(file)
    open(con)
    on.exit({
         close(con)
    })
    
    probes <- read.delim(con, sep="\t", header=TRUE, skip=probesHead, 
                       nrows=nProbes, quote="", stringsAsFactors=FALSE)
    controls <- read.delim(con, sep="\t", header=TRUE, skip = 1, 
                         nrows=nControls, quote="", stringsAsFactors=FALSE)
  
    list(probes = probes, controls = controls)
}

## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## System tests for readBPM()
##
## Data is automatically downloaded from Illumina (e.g. [1]), if missing.
##
## NOTE: These tests will only run if '_R_CHECK_FULL_' is set, i.e.
## they won't run on Bioconductor or CRAN servers.
##
## REFERENCES:
## [1] http://support.illumina.com/downloads/infinium_bovineld_beadchip_product_files.ilmn
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if (FALSE && Sys.getenv("_R_CHECK_FULL_") != "") {

## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## Utility functions
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
getIlluminaAnnotationFile <- function(chipType, filename, force=FALSE) {
  # Share data files across architectures, e.g. tests_i386 and tests_x64.
  rootPath <- "../annotationData,illuminaio,tests"
  path <- file.path(rootPath, "chipTypes", chipType)
  if (!file_test("-d", path)) dir.create(path, recursive=TRUE)
  pathname <- file.path(path, filename)

  # Already downloaded?
  if (!force && file_test("-f", pathname)) return(pathname)

  # Download from Illumina
  urlRoot <- "http://supportres.illumina.com/documents/downloads/productfiles"
  url <- file.path(urlRoot, chipType, filename)
  download.file(url, destfile=pathname, mode="wb")
  if (!file_test("-f", pathname)) stop("Failed to download file: ", url)

  pathname
} # getIlluminaAnnotationFile()


## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## Tests
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## pathname <- getIlluminaAnnotationFile("bovineld", "bovineld_c.bpm")

path <- "annotationData/chipTypes/HumanMethylation450"
filename <- "humanmethylation450_15017482_v-1-2.bpm"
pathname <- file.path(path,filename)
data <- illuminaio::readBPM(pathname)
str(data)

} # if (Sys.getenv("_R_CHECK_FULL_") != "")

readIDAT_nonenc <- function(file, what = c("all", "IlluminaID", "nSNPsRead")) {
    readByte <- function(con, n=1, ...) {
        readBin(con, what="integer", n=n, size=1, endian="little", signed=FALSE)
    }

    readShort <- function(con, n=1, ...) {
        readBin(con, what="integer", n=n, size=2, endian="little", signed=FALSE)
    }

    readInt <- function(con, n=1, ...) {
        readBin(con, what="integer", n=n, size=4, endian="little", signed=TRUE)
    }

    readLong <- function(con, n=1, ...) {
        readBin(con, what="integer", n=n, size=8, endian="little", signed=TRUE)
    }

    readBytesToRead <- function(con, ...) {
        m <- readByte(con, n=1)
        n <- m %% 128
        shift <- 0L
        while (m %/% 128 == 1) {
            ## Read next length byte ...
            m <- readByte(con, n=1)
            ## ... which represents the next 7 hi-bits
            shift <- shift + 7L
            k <- (m %% 128) * 2^shift
            ## Total number of bytes to read
            n <- n + k
        }

        ## 'n' is a numeric; not an integer
        n
    }

    readString <- function(con, ...) {
        ## From [1] https://code.google.com/p/glu-genetics/source/browse/glu/lib/illumina.py#86:
        ## String data are encoded as a sequence of one or more length
        ## bytes followed by the specified number of data bytes.
        ##
        ## The lower 7 bits of each length byte encodes the bits that
        ## comprise the length of the following byte string.  When the
        ## most significant bit it set, then an additional length byte
        ## follows with 7 additional high bits to be added to the current
        ## length.  The following string lengths are accommodated by
        ## increasing sequences of length bytes:
        ##
        ## length  maximum
        ## bytes   length
        ## ------  --------
        ##   1       127 B
        ##   2        16 KB
        ##   3         2 MB
        ##   4       256 MB
        ##   5        32 GB
        ##
        ## While this seems like a sensible progression, there is some
        ## uncertainty about this interpretation, since the longest of
        ## string observed in the wild has been of length 6,264 with
        ## two length bytes.
        ##
        ## EXAMPLES: by HB (2015-09-11)
        ## Length   Len bytes  Iterations (n, m, k, shift)
        ## 1        (1)
        ## 127      (127)       -> n=128
        ## 128      (128,1)     -> n=0  ,m=1  ,shift=7, k=128 -> n=128
        ## 255      (255,1)     -> n=128,m=1  ,shift=7, k=128 -> n=255
        ## 256      (128,2)     -> n=0  ,m=2  ,shift=7, k=256 -> n=256
        ## 257      (129,2)     -> n=1  ,m=2  ,shift=7, k=256 -> n=257
        ## 512      (128,4)     -> n=0  ,m=4  ,shift=7, k=512 -> n=512
        ## 81921    (129,128,5) -> n=1  ,m=128,shift=7, k=0 -> n=1+0=1
        ##                      -> n=0  ,m=5  ,shift=14,k=81920
        ##                                              -> n=1+81920=81921

        ## Parse the number of characters to read
        n <- readBytesToRead(con)

        ## Read the data bytes
        readChar(con, nchars=n)
    }

    readUnknownBytes <- function(con, ...) {
        ## Parse the number of characters to read
        n <- readBytesToRead(con)
        
        ## Read the data bytes
        c(as.integer(n), readByte(con, n = n))
    }

    readField <- function(con, field) {
        switch(field,
               "IlluminaID" = readInt(con = con, n=nSNPsRead),
               "SD" = readShort(con = con, n=nSNPsRead),
               "Mean"= readShort(con = con, n=nSNPsRead),
               "NBeads" = readByte(con = con, n=nSNPsRead),
               "MidBlock" = {
                   nMidBlockEntries <- readInt(con = con, n=1)
                   MidBlock <- readInt(con = con, n=nMidBlockEntries)
               },
               "RedGreen" = readInt(con = con, n=1),
               "MostlyNull" = readString(con = con),
               "Barcode" = readString(con = con),
               "ChipType" = readString(con = con),
               "MostlyA" = readString(con = con),
               "Unknown.1" = readString(con = con),
               "Unknown.2" = readString(con = con),
               "Unknown.3" = readString(con = con),
               "Unknown.4" = readString(con = con),
               "Unknown.5" = readString(con = con),
               ## We don't know what 'Unknown.6' should be, but we know of at
               ## least one instance [2] where using readString() would trigger
               ## a warning about nulls in the string stemming from the byte
               ## sequence (1, 0). Because of this, we choose to read it as
               ## a byte sequence based of readUnknownBytes(), because it's
               ## the closest to readString() which we have used for years.
               ## We might change how this field is parsed in a future version
               ## when better understood what this field contains, e.g.
               ## it could be that readShort() should be used. /HB 2022-09-23
               ## [2] https://github.com/HenrikBengtsson/illuminaio/issues/21
               "Unknown.6" = readUnknownBytes(con = con),
               "Unknown.7" = readString(con = con),
               "RunInfo" = {
                   nRunInfoBlocks <- readInt(con = con, n=1)
                   naValue <- as.character(NA)
                   RunInfo <- matrix(naValue, nrow=nRunInfoBlocks, ncol=5)
                   colnames(RunInfo) <- c("RunTime", "BlockType", "BlockPars",
                                          "BlockCode", "CodeVersion")
                   for (ii in seq_len(nRunInfoBlocks)) {
                       for (jj in 1:5) {
                           RunInfo[ii,jj] <- readString(con = con)
                       }
                   }
                   RunInfo
               },
               stop("readIDAT_nonenc: unknown field"))
    }

    if(! (is.character(file) || try(isOpen(file))))
        stop("argument 'file' needs to be either a character or an open, seekable connection")
    what <- match.arg(what)

    if(is.character(file)) {
        stopifnot(length(file) == 1)
        file <- path.expand(file)
        stopifnot(file.exists(file))
        fileSize <- file.info(file)$size
        if(grepl("\\.gz$", file))
            con <- gzfile(file, "rb")
        else
            con <- file(file, "rb")
        on.exit({
            close(con)
        })
    } else {
        con <- file
        fileSize <- 0
    }

    if(!isSeekable(con))
        stop("The file connection needs to be seekable")

    ## Assert file format
    magic <- readChar(con, nchars=4)
    if (magic != "IDAT") {
        stop("Cannot read IDAT file. File format error. Unknown magic: ", magic)
    }

    ## Read IDAT file format version
    version <- readLong(con, n=1)
    if (version < 3) {
        stop("Cannot read IDAT file. Unsupported IDAT file format version: ", version)
    }

    ## Number of fields
    nFields <- readInt(con, n=1)

    fields <- matrix(0L, nrow=nFields, ncol=3)
    colnames(fields) <- c("fieldCode", "byteOffset", "Bytes")
    for (ii in 1:nFields) {
        fields[ii,"fieldCode"] <- readShort(con, n=1)
        fields[ii,"byteOffset"] <- readLong(con, n=1)
    }

    ## The below stems from [1].  However, there is now also [3], which seems to
    ## have identified a few more of the 'Unknowns' fields. /HB 2022-09-23
    ## [3] https://github.com/bioinformed/glu-genetics/blob/dcbbbf67a308d35e157b20a9c76373530510379a/glu/lib/illumina.py#L44-L61
    knownCodes <- c(
        "nSNPsRead"  = 1000,
        "IlluminaID" =  102,
        "SD"         =  103,
        "Mean"       =  104,
        "NBeads"     =  107,
        "MidBlock"   =  200,
        "RunInfo"    =  300,
        "RedGreen"   =  400,
        "MostlyNull" =  401, # 'Manifest' [1,3]
        "Barcode"    =  402,
        "ChipType"   =  403,
        "MostlyA"    =  404, # 'Stripe' [1], 'label' [3]
        "Unknown.1"  =  405, # 'opa' [3]
        "Unknown.2"  =  406, # 'Sample ID' [1,3]
        "Unknown.3"  =  407, # 'descr' [3]
        "Unknown.4"  =  408, # 'Plate' [1,3]
        "Unknown.5"  =  409, # 'Well' [1,3]
        "Unknown.6"  =  410, 
        "Unknown.7"  =  510  # 'unknown' [1,3]
        )

    nNewFields <- 1
    rownames(fields) <- paste("Null", 1:nFields)
    for (ii in 1:nFields) {
        temp <- match(fields[ii,"fieldCode"], knownCodes)
        if (!is.na(temp)) {
            rownames(fields)[ii] <- names(knownCodes)[temp]
        } else {
            rownames(fields)[ii] <- paste("newField", nNewFields, sep=".")
            nNewFields <- nNewFields + 1
        }
    }

    stopifnot(min(fields[, "byteOffset"]) == fields["nSNPsRead", "byteOffset"])

    seek(con, where=fields["nSNPsRead", "byteOffset"], origin="start")
    nSNPsRead <- readInt(con, n=1)
    if(what == "nSNPsRead")
        return(nSNPsRead)
    if(what == "IlluminaID") {
        where <- fields["IlluminaID", "byteOffset"]
        seek(con, where = where, origin = "start")
        res <- readField(con = con, field = "IlluminaID")
        return(as.character(res))
    }

    res <- rownames(fields)
    names(res) <- res
    res <- res[order(fields[res, "byteOffset"])]
    res <- res[names(res) != "nSNPsRead"]
    res <- lapply(res, function(xx) {
        where <- fields[xx, "byteOffset"]
        seek(con, where = where, origin = "start")
        readField(con = con, field = xx)
    })

    Unknowns <- list(
      MostlyNull=res$MostlyNull,
      MostlyA=res$MostlyA
    )
    names <- grep("^Unknown[.]", names(res), value = TRUE)
    Unknowns[names] <- res[names]

    Quants <- cbind(res$Mean, res$SD, res$NBeads)
    colnames(Quants) <- c("Mean", "SD", "NBeads")
    rownames(Quants) <- as.character(res$IlluminaID)

    list(
        fileSize=fileSize,
        versionNumber=version,
        nFields=nFields,
        fields=fields,
        nSNPsRead=nSNPsRead,
        Quants=Quants,
        MidBlock=res$MidBlock,
        RedGreen=res$RedGreen,
        Barcode=res$Barcode,
        ChipType=res$ChipType,
        RunInfo=res$RunInfo,
        Unknowns=Unknowns
    )
}

%% \VignetteIndexEntry{Introduction to illuminaio}
\documentclass{article}

<<style, echo=FALSE, results=tex>>=
BiocStyle::latex()
@

\title{illuminaio}
\author{M.L. Smith, K.D. Hansen}

\begin{document}

\maketitle

\section*{Introduction}

\Biocpkg{illuminaio} is designed to provide a single package containing routines for importing data from Illumina BeadArray platforms into \R{}.  The intention is for \Biocpkg{illuminaio} to provide developers of downstream analysis packages with a mechanism for extracting all possible information from IDAT files in a relatively straightforward fashion.  The choice of data to retain and how it should be stored is then left for the end user to decide.

This vignette gives examples of how data files can be read and discusses the values that are returned, which will vary depending upon the BeadArray platform used.  It also demonstrates that the values extracted by \Biocpkg{illuminaio} are the same as those returned when using Illumina's own GenomeStudio software.

\section*{Citing illuminaio}

If you use the package, please cite the associated paper \cite{illuminaio}.

\section*{Reading Data}

\subsection*{Expression Array}

The code below gives an example of how an IDAT file (in this case from a Human expression array) can be read, and explores the information that is extracted.

<<loadData>>=
library(illuminaio)
library(IlluminaDataTestFiles)
idatFile <- system.file("extdata", "idat", "4343238080_A_Grn.idat",
                        package = "IlluminaDataTestFiles")
idat <- readIDAT(idatFile)                                 
@ 

The first two lines above load libraries require for this vignette.  Firstly this package and then \Biocpkg{IlluminaDataTestFiles}, a small data package containing the example files that are used throughout this vignette.  The third line generates the complete path to one such file.  We then use the function \Rfunction{readIDAT} to read the file.  This function take only a single argument, the file's path.  Although IDAT files are found in multiple formats, \Rfunction{readIDAT} is able to determine this and will call the appropriate reading routine internally.  The returned object is a list, the contents of which is explored below.

<<exploreData>>=
names(idat)
@ 

Using the \Rfunction{names} command above lists the data extracted from the file.  In this case \texttt{Barcode} and \texttt{Section} are the identifiers for the BeadChip and can usually be found in the file name as well.  \texttt{ChipType} describes the BeadArray platform this file was generated from.  The \texttt{RunInfo} slot holds information about the processing performed on the array, most notably the date upon which it was scanned.  Such information may be useful if one is attempting to identify batch effects amongst multiple samples.  \texttt{Quants} is where the per-bead-type values are found.  The commands below assign the \texttt{Quants} values to a new variable for convenience, and then print the first six entries from the resulting data.frame.

<<printQuants>>=
idatData <- idat$Quants    
head(idatData)
@ 

We can see that for this expression array a total of 10 values are returned for each bead-type.  The column headers are pulled directly from the IDAT file and do not directly match with how things are commonly labeled in GenomeStudio, although for the most part they are relatively easy to decipher.  The \texttt{CodeBinData, MeanBinData} and \texttt{NumGoodBeadsBinData} columns are those that are reported by default in GenomeStudio and correspond the ProbeID, AVG\_Signal and NBEADS values respectively.  \texttt{DevBinData} gives the standard deviation for the bead-type and can be used the generate the BEAD\_STDERR values GenomeStudio reports.  The remaining columns give additional information and are discussed in the file \texttt{EncryptedFormat.pdf} that also accompanies this package.

\subsection*{Genotyping Array}

The example above focused on reading an IDAT file produced by scanning an expression array.  To highlight some of the differences in output we shall now read a file from an Infinium genotyping array.

\fixme{I should check exactly which platform this is from - MLS}

<<readGenotyping>>=
genotypeIdatFile <- system.file("extdata", "idat", "5723646052_R02C02_Grn.idat",
                        package = "IlluminaDataTestFiles")
genotypeIdat <- readIDAT(genotypeIdatFile)
names(genotypeIdat)
@ 

The reading of the file proceeds in exactly the same way as before and a list is again returned.  However, there are several more data fields returned.  Again \texttt{Quants} is where the per-bead-type values are stored.

<<printGenotypingQuants>>=
head(genotypeIdat$Quants)
@ 

For genotyping arrays only the four typically reported values are contained within the IDAT file and their column names more closely resemble those that are found in GenomeStudio.

\section*{Comparison with GenomeStudio}

Now we shall compare the values extracted by \Biocpkg{illuminaio} with those reported by Illumina's GenomeStudio software, to ensure our file reading routines are performing correctly. 

\subsection*{Importing GenomeStudio Values}

<<ImportGenomeStudio>>=
gsFile <- system.file("extdata", "gs", "4343238080_A_ProbeSummary.txt.gz",
                      package = "IlluminaDataTestFiles")
gStudio <- read.delim(gsFile, sep = "\t", header = TRUE)
idatData <- idatData[which(idatData[,"CodesBinData"] %in% gStudio[,"ProbeID"]),]
gStudio <- gStudio[match(idatData[,"CodesBinData"], gStudio[,"ProbeID"]),]
@ 

The first line above reads a file that was produced by reading the IDAT file into GenomeStudio and then immediately exporting that data as a tab separated text file.  No other processing was performed on the data.  \fixme{Currently this file is stored on a web server, although the intention is to include it in the \Biocpkg{illuminaDataTestFile} package.}

However, the two datasets are not quite compatible at the moment.  Reading directly from an IDAT file returns values for several bead-types that serve as internal controls and are not annotated by Illumina.  These bead-types are excluded automatically by GenomeStudio, so the second line above identifies and removes them from our \Biocpkg{illuminaio} data.  The two data sets should now contain the same number of bead-types.  

The inclusion of these extra bead-types is not the only difference, they are also in different orders. Bead-types are extracted in numerical order from IDAT files, but the GenomeStudio output is sorted alphabetically.  The third line reorders the GenomeStudio values to match those from \Biocpkg{illuminaio}, making our comparison slightly easier.

\subsection*{Performing Comparison}

The code below produces the two plots seen in Figure 1.

<<figureComparingValues, fig=TRUE, include=FALSE, width=10, height=5.2, png=TRUE, pdf=FALSE>>=
par(mfrow = c(1,2))
plot(idatData[, "MeanBinData"], gStudio[, "X4343238080_A.AVG_Signal"], 
     xlab = "illuminaio", ylab = "GenomeStudio")
identical(idatData[, "MeanBinData"], gStudio[, "X4343238080_A.AVG_Signal"])
hist(idatData[, "MeanBinData"]- gStudio[, "X4343238080_A.AVG_Signal"],
     breaks = 100, main = "", xlab = "Difference")
@ 
\incfig{illuminaio-figureComparingValues}{\textwidth}{}{Comparing values obtained by \Biocpkg{illuminaio} and GenomeStudio}

The first plot shows the summarized bead-intensity values extracted by \Biocpkg{illuminaio} on the horizontal axis against GenomeStudio's values on the vertical axis.  We can see they are highly similar.  However, they are not identical, as shown by the third line above.  The second plot visualizes the distribution of the differences between the two sets of values, showing them to be small.  These are most likely introduced by rounding performed by GenomeStudio that is not carried out by \Biocpkg{illuminaio}.



\pagebreak



%---------------------------------------------------------
\section*{Session info}
%---------------------------------------------------------
Here is the output of \Rfunction{sessionInfo} on the system on which
this document was compiled:
<<sessionInfo, results=tex, print=TRUE, echo=FALSE>>=
toLatex(sessionInfo())
@

\bibliography{illuminaio}

\end{document}

\name{readBGX}
\alias{readBGX}

\title{Low-level function to read BGX files}

\description{
    Reads BGX files (array design information) from Illumina expression arrays.
}

\usage{
readBGX(file)
}

\arguments{
  \item{file}{character string specifying BGX file to be read in}
}

\value{
  A list.
}

\details{
  Illumina provide probe annotation for expression arrays as either tab separated text files or as files with the extension \file{.bgx}.  BGX files are in fact just gzipped versions of the text files.  This function reads such files and returns a list with two entries, the first containing the target-probe information and the second containing details of the control probes.
}

\author{
  Mike Smith.  Based loosely on code developed by Mark Dunning available in \pkg{beadarray}~1.16.0.
}
\keyword{IO}

\name{readIDAT}
\alias{readIDAT}

\title{Low-level function to read IDAT files}

\description{ Reads intensity information for each bead type from a
  single IDAT file for Infinium II, Infinium HD and expression platforms.
}

\usage{
readIDAT(file, what = c("all", "IlluminaID", "nSNPsRead"))
}

\arguments{
  \item{file}{character string specifying IDAT file to be read in.}
  \item{what}{This allows the return of parts of the IDAT file, see
  Value. This argument is only supported for non-encrypted IDAT files;
  setting it to be different from default on an encrypted file throws an
  error.}
}

\details{
  This function returns a list containing summarized intensities and other
  information extracted from a single IDAT file.

  Thanks to Keith Baggerly for providing the first version of this code.

  There are two commonly encountered versions of IDAT.  One version uses
  encryption to store data and is the format usually produced by Illumina
  expression arrays.  The other version is unencrypted files and
  is found in data from SNP arrays and DNA methylation arrays.

  This function is able to handle data in either format, scanning the
  file header and calling the appropriate reading routine internally.

  The function supports reading gzipped, unencrypted IDAT files.

  The use of the \code{what} argument is for package writers.
}

\value{
  The exact data returned will vary depending upon whether the input
  file was encrypted or not.  In either case, everything that can be
  extract from the file will be returned in the form of a list.
    
  For both file types the most important entry in the returned list is the
  item \code{Quants}.  When reading unencrypted files this contains average intensity (\code{Mean}),
  number of beads (\code{NBeads}) and a measure of variability
  (\code{SD}) for each bead type on the array.  For expression arrays
  in addition to these some additional information is available,
  including median and trimmed-mean estimates of average intensity,
  averaged local background intensities and the number of beads present
  before outliers where excluded.

  For both file types details of scanning and decoding dates are
  included in the \code{RunInfo} slot.
  
  A further list entry, \code{Extra}, may also be present.  Occasionally an field
  in the IDAT file will have a number of data points inconsistent with the remainder
  of the file and as such is not returned in the \code{Quants} entry.  Such
  fields will be placed here.  So far only the HumanHap550 v1 BeadChip has exhibited 
  this property.

  In case \code{what} is set to \code{nSNPsRead} the function returns an
  integer equal to this field.  In case \code{what} is set to
  \code{IlluminaID} the function returns a character vector containing
  the IDs.
}

\references{
  Ritchie ME, Carvalho BS, Hetrick KN, Tavar\'e S, Irizarry RA.
  R/Bioconductor software for Illumina's Infinium whole-genome
  genotyping BeadChips. Bioinformatics. 2009 Oct 1;25(19):2621-3.
}

\author{Unencrypted: Keith Baggerly, with modifications by Matt Ritchie, Henrik
  Bengtsson and Kasper Daniel Hansen.
Encrypted: Mike Smith}

\examples{
if(require(IlluminaDataTestFiles)) {
  idatFile <- system.file("extdata", "idat", "4019585376_B_Red.idat",
                          package = "IlluminaDataTestFiles")
  idat <- readIDAT(idatFile)
  names(idat)
  idat$Quants[1:5,]
}
}
\keyword{IO}

\name{readBPM}
\alias{readBPM}
\alias{readBPM}

\title{Low-level function to read BPM files}

\description{
  Reads BPM files (array design information) from Illumina arrays.

  \emph{WARNING: This function is deprecated since illuminaio 0.1.4
  (January 2016), because its parser is invalid/incorrect. (Issue #6)}
}

\usage{
readBPM(file)
}

\arguments{
  \item{file}{character string specifying BPM file to be read in}
}

\value{
  A list.
}

\author{
  Originally created by Keith Baggerly in 2008.
  Later included in the \pkg{crlmm} package.
}
\keyword{IO}
\keyword{internal}
